import logging
from dockwrkr.monads import *
from dockwrkr import (Command)
import tabulate

logger = logging.getLogger(__name__)

class Status(Command):

    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "dockwrkr status [options] CONTAINER..."

    def getHelpTitle(self):
        return "Output the container status table"

    def main(self):
        return self.core.status(self.args) \
            .bind(self.tabulateStatus) \
            .catch(self.exitError) \
            .bind(logger.info)

    def tabulateStatus(self, containerStatuses):
        logger.debug("STATUSES: %s" %containerStatuses)

        headers = ["NAME", "CONTAINER", "PID", "IP", "UPTIME", "EXIT"]
        tabulate.MIN_PADDING = 8
        table = tabulate.tabulate(
            containerStatuses, headers=headers, tablefmt="plain")
        return OK(table)



from dockwrkr import (Command)
from dockwrkr.console import PassThroughParser


class Exec(Command):

    def getParserClass(self):
        return PassThroughParser

    def getShellOptions(self, optparser):
        optparser.add_option("-t", "--tty", dest="tty",
                             help="Allocate a pseudo-TTY", action="store_true")
        optparser.add_option("-i", "--interactive", dest="interactive",
                             help="Keep STDIN open even if not attached", action="store_true")
        optparser.add_option("--privileged", dest="privileged",
                             help="Give extended privileges to the command", action="store_true")
        optparser.add_option("-d", "--detach", dest="detach",
                             help="Run in the background", action="store_true")
        optparser.add_option("-u", "--user", dest="user",
                             help="Username or UID")
        return optparser

    def getUsage(self):
        return "dockwrkr exec [options] CONTAINER"

    def getHelpTitle(self):
        return "Run a command in a running container"

    def main(self):
        containers = self.args
        if not len(self.args) > 0:
            return self.exitWithHelp("Please provide a container.")
        if not len(self.args) > 1:
            return self.exitWithHelp("Please provide a command.")
        container = self.args[0]
        command = self.args[1:]
        return self.core.excmd(
            container=container,
            cmd=command,
            tty=self.getOption('tty'),
            user=self.getOption('user'),
            detach=self.getOption('detach'),
            privileged=self.getOption('privileged'),
            interactive=self.getOption('interactive')
        ).catch(self.exitError)




class DockwrkrError(Exception):
    '''
    Base class to all dockwrkr exceptions
    '''
    errorLabel = ""

    def __init__(self, message=''):
        super(DockwrkrError, self).__init__(message)
        self.errorLabel = message


class ConfigFileNotFound(DockwrkrError):
    ''' Config file not found '''


class ConfigSyntaxError(DockwrkrError):
    ''' Config syntax error '''


class FileSystemError(DockwrkrError):
    ''' File system exception '''


class FileDoesNotExist(DockwrkrError):
    ''' File error exception '''


class ShellCommandError(DockwrkrError):
    def __init__(self, code=None, message='', stdout='', stderr='', cmd=''):
        super(ShellCommandError, self).__init__(message)
        self.code = code
        self.stdout = stdout
        self.stderr = stderr
        self.cmd = cmd


class UserInterruptError(ShellCommandError):
    ''' Raised when the user interrupts a process '''


class InvalidCommandError(DockwrkrError):
    '''
    Raised when an invalid command is passed to CLI
    '''


class InvalidOptionError(DockwrkrError):
    '''
    Raised when an invalid option is passed to CLI
    '''


class InvalidConfigError(DockwrkrError):
    '''
    Raised on an invalid configuration is found
    '''


class InvalidContainerError(DockwrkrError):
    ''' Invalid container was specified '''


class InvalidRegistryError(DockwrkrError):
    ''' Invalid registry was specified '''


class DockerError(ShellCommandError):
    '''
    Raised when a docker error is encountered
    '''



import os
import errno
import sys
import logging
import collections
import shutil
import hashlib
from time import time
from datetime import datetime
from math import floor
from pkg_resources import Requirement, resource_filename, require
import yaml

try:  # py3
    from shlex import quote
except ImportError:  # py2
    from pipes import quote

from dockwrkr.exceptions import (
    ConfigSyntaxError,
    FileSystemError,
    FileDoesNotExist
)

logger = logging.getLogger(__name__)

_yaml_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def _dict_representer(dumper, data):
    return dumper.represent_dict(iter(data.items()))


def _dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))


yaml.add_representer(collections.OrderedDict, _dict_representer)
yaml.add_constructor(_yaml_mapping_tag, _dict_constructor)


def getPackageVersion():
    return require('dockwrkr')[0].version


def ensureList(v):
    if v:
        return [x for x in (v if isinstance(v, (list, tuple)) else [v])]
    else:
        return []


def expandLocalPath(path, basePath):
    if not os.path.isabs(path):
        path = os.path.normpath(path)
        path = os.path.join(basePath, path)
    return os.path.abspath(os.path.expanduser(os.path.realpath(path)))


def walkUpForFile(root, findfile):
    lastRoot = root
    while root:
        for base, dirs, files in os.walk(root):
            del dirs[0:len(dirs)]
            if findfile in files:
                return os.path.join(base, findfile)
        lastRoot = root
        root = os.path.dirname(lastRoot)
        if root == lastRoot:
            break


def dateToAgo(time=False):
    """
    Get a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc
    """
    now = datetime.now()
    if type(time) is int:
        diff = now - datetime.fromtimestamp(time)
    elif isinstance(time, datetime):
        diff = now - time
    elif not time:
        diff = now - now
    second_diff = diff.seconds
    day_diff = diff.days

    if day_diff < 0:
        return ''

    if day_diff == 0:
        if second_diff < 10:
            return "just now"
        if second_diff < 60:
            return str(second_diff) + " seconds ago"
        if second_diff < 120:
            return "a minute ago"
        if second_diff < 3600:
            return str(floor(second_diff / 60)) + " minutes ago"
        if second_diff < 7200:
            return "an hour ago"
        if second_diff < 86400:
            return str(floor(second_diff / 3600)) + " hours ago"
    if day_diff == 1:
        return "Yesterday"
    if day_diff < 7:
        return str(day_diff) + " days ago"
    if day_diff < 31:
        return str(floor(day_diff / 7)) + " weeks ago"
    if day_diff < 365:
        return str(floor(day_diff / 30)) + " months ago"
    return str(floor(day_diff / 365)) + " years ago"


def writeToFile(data, filename):
    try:
        with open(filename, "wb") as fh:
            fh.write(data)
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def writeYAML(filename, data):
    try:
        with open(filename, "w") as fileh:
            fileh.write(yaml.dump(data, default_flow_style=False))
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def readYAML(filename):
    try:
        stream = open(filename, "r")
        contents = yaml.load(stream)
        return contents
    except yaml.YAMLError as exc:
        msg = "Syntax error in file %s"
        if hasattr(exc, 'problem_mark'):
            #pylint: disable=E1101
            mark = exc.problem_mark
            msg += " Error position: (%s:%s)" % (mark.line +
                                                 1, mark.column + 1)
        raise ConfigSyntaxError(msg)
    except Exception as err:
        raise FileSystemError(
            "Failed to read configuration file %s : %s" % (filename, err))


def humanReadableBytes(num):
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
        return "%3.1f%s" % (num, 'TB')


def sha1sum(filename):
    sha = hashlib.sha1()
    bufsize = 65536
    with open(filename, 'rb') as fd:
        while True:
            data = fd.read(bufsize)
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


def pathComponents(path):
    folders = []
    path = os.path.normpath(path)
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)
            break

    folders.reverse()
    return folders


def mergeDict(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                mergeDict(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def safeQuote(string):
    return quote(str(string))




import os
import sys
import logging
from dockwrkr import (Program, Core)
from dockwrkr.utils import getPackageVersion
import dockwrkr.logs

logger = logging.getLogger(__name__)

class DockwrkrCLI(Program):
    """Dockwrkr CLI command"""

    def __init__(self):
        super(DockwrkrCLI, self).__init__()
        self.name = 'dockwrkr'

    def setupCommands(self):
        self.addCommand('help', 'dockwrkr.command.help')
        self.addCommand('create', 'dockwrkr.command.create')
        self.addCommand('start', 'dockwrkr.command.start')
        self.addCommand('stop', 'dockwrkr.command.stop')
        self.addCommand('remove', 'dockwrkr.command.remove')
        self.addCommand('restart', 'dockwrkr.command.restart')
        self.addCommand('recreate', 'dockwrkr.command.recreate')
        self.addCommand('reset', 'dockwrkr.command.reset')
        self.addCommand('pull', 'dockwrkr.command.pull')
        self.addCommand('status', 'dockwrkr.command.status')
        self.addCommand('exec', 'dockwrkr.command.exec')
        self.addCommand('stats', 'dockwrkr.command.stats')
        self.addCommand('login', 'dockwrkr.command.login')
        self.addCommand('run', 'dockwrkr.command.run')
        return self

    def getShellOptions(self, optparser):
        optparser.add_option("-f", dest="configFile",
                             help="Override default config file")
        optparser.add_option(
            "-d", dest="debug", help="Activate debugging output", default=False, action="store_true")
        optparser.add_option(
            "-y", dest="assumeYes", help="Assume yes when prompted", default=False, action="store_true")
        return optparser

    def getUsage(self):
        return "dockwrkr [options] COMMAND [command-options]"

    def getHelpTitle(self):
        version = getPackageVersion()
        return "Docker container composition (version: %s)" % version

    def initCommand(self, command):
        core = Core()
        if self.getOption('assumeYes'):
            command.assumeYes = True

        if self.getOption('configFile') and os.path.isfile(self.getOption('configFile')):
            logger.debug("Configuration file: %s" % self.getOption('configFile'))
            core.configFile = self.getOption('configFile')

        if command.autoInitCore:
            core.initialize().catch(self.exitError)

        command.core = core
        return command


def cli():
    args = sys.argv
    args.pop(0)

    prog = DockwrkrCLI()
    prog.execute(args)



from dockwrkr import (Command)


class Stats(Command):

    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "dockwrkr stats [options] CONTAINER..."

    def getHelpTitle(self):
        return "Output live stats for the listed containers"

    def main(self):
        containers = self.args
        return self.core.stats(self.args) \
            .catch(self.exitError)



import sys
import logging
import logging.config
from dockwrkr.monads import *


class StdoutFilter(logging.Filter):
    def __init__(self, level):
        self._level = level
        logging.Filter.__init__(self)

    def filter(self, rec):
        return rec.levelno < self._level


class LevelLogFormatter(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.DEBUG:
            record.msg = '[%s] %s' % (record.levelname, record.msg)
        return super(LevelLogFormatter, self).format(record)


LOG_SETTINGS = {
    'version': 1,
    'disable_existing_loggers': False,
    'root': {
        'level': 'NOTSET',
    },
    'loggers': {
        'dockwrkr': {
            'level': 'NOTSET',
            'qualname': 'dockwrkr',
            'handlers': ['stdout', 'stderr'],
            'propagate': False
        }
    },
    'handlers': {
        'console': {
            'level': 'NOTSET',
            'formatter': 'levelformatter',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
        },
        'stdout': {
            'level': 'INFO',
            'formatter': 'levelformatter',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
            'filters': ['StdoutFilter'],
        },
        'stderr': {
            'level': 'WARNING',
            'formatter': 'levelformatter',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr'
        }
    },
    'filters': {
        'StdoutFilter': {
            '()': StdoutFilter,
            'level': logging.WARNING
        }
    },
    'formatters': {
        'levelformatter': {
            '()': LevelLogFormatter, 'format': '%(message)s'},
        'stdout': {'format': '%(message)s'},
        'stderr': {'format': '%(message)s'}
    }
}

logging.config.dictConfig(LOG_SETTINGS)

logger = logging.getLogger("dockwrkr")


def ddebug(msg, *args, **kwargs):
    return dlog(logging.DEBUG, msg, *args, **kwargs)


def dinfo(msg, *args, **kwargs):
    return dlog(logging.INFO, msg, *args, **kwargs)


def dwarning(msg, *args, **kwargs):
    return dlog(logging.WARNING, msg, *args, **kwargs)


def dcritical(msg, *args, **kwargs):
    return dlog(logging.CRITICAL, msg, *args, **kwargs)


def dexception(msg, *args, **kwargs):
    return dlog(logging.EXCEPTION, msg, *args, **kwargs)


def dlog(level, msg, *args, **kwargs):
    if args:
        msg = msg % args

    def deferredLog(data=None, *args, **kwargs):
        logger.log(level, msg)
        return OK(data)
    return deferredLog
